/*************************************************************************\
* Copyright (c) 2000-2004 The University of Chicago, as Operator of Argonne
* National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// CCaObject2.cpp : Implementation of CCaObject2 class

#define CCAOBJECT2_EXPORTS

#define UNREFERENCED(x) (x)
#define ERRMSG if(errHandler) (*errHandler)

#ifdef WIN32
#include <Windows.h>
#endif
#include <limits.h>
#include <float.h>
#ifndef UNDER_CE
// WCE doesn't have time.h
#include <time.h>
#endif
#define MAX_CONVERT_SIZE 64

// Define TRUE and FALSE
#if !defined(TRUE) || ((TRUE) != 1)
#define TRUE (1)
#endif
#if !defined(FALSE) || ((FALSE) != 0)
#define FALSE (0)
#endif
#if !defined(ERR) || ((ERR) != -1)
#define ERR (-1)
#endif
#if !defined(OK) || ((OK) != 0)
#define OK (0)
#endif

// Define MAX and MIN
#ifndef MAX
#define MAX(a,b)  ((a)>(b)?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b)  ((a)<(b)?(a):(b))
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cvtFast.h>

#include "CCaObject2.h"

// Function prototypes
static int localCvtLongToHexString(long source, char  *pdest);
static void localCvtDoubleToExpNotationString(double value, char *textField,
  unsigned short precision);
static void localCvtDoubleToString(double flt_value, char  *pstr_value,
  unsigned short precision);

// extern "C" wrappers needed by CA routines for internal callbacks
extern "C" {
    static void accessRightsICB(struct access_rights_handler_args args);
    static void connectionICB(struct connection_handler_args args);
    static void exceptionICB(struct exception_handler_args args);
#ifdef todo
    static void fdRegistrationICB(void *dummy, int fd, int condition);
#endif
    static void graphicalInfoICB(struct event_handler_args args);
    static void valueChangedICB(struct event_handler_args args);
}

// Static initializations
CCaObject2ExceptionCB CCaObject2::exceptionCB;
CCaObject2ErrorHandler CCaObject2::errHandler;
int CCaObject2::caStarted=0;
unsigned long CCaObject2::refCount=0;

/***************************** Constructor ****************************/
CCaObject2::CCaObject2()
{
    initialize();

    if(!caStarted) taskInit();
    refCount++;
}

/***************************** Destructor *****************************/
CCaObject2::~CCaObject2()
{
    if(connected || Chid) disconnect();
    if(data) delete [] data;
    data=NULL;
    refCount--;
    if(refCount <= 0) taskExit();
}

/***************************** initialize *****************************/
void CCaObject2::initialize(void)
{
    int i;

    connected=0;
    monitored=monitor_off;
    elementCount=0;
    dataType=0;
    value=0.0;
    hopr=0.0;
    lopr=0.0;
    precision=0;
    status=0;;
    severity=0;
    pvCurConnected=0;
    pvPrevConnected=0;
    readAccess=0;
    writeAccess=0;
    graphicalInfoAvailable=0;
    for(i=0; i < MAX_ENUM_STATES; i++) stateStrings[i]=NULL;
    *name='\0';
    array=NULL;
    time.secPastEpoch=time.nsec=0;
    monitorStatusChanged=0;
    monitorSeverityChanged=0;
    monitorValueChanged=0;
    monitorZeroTransition=0;
    data=NULL;
    dataSize=0;
#ifdef todo
    info=;
#endif
    Chid=NULL;
    Evid=NULL;
    
    accessRightsCB=NULL;
    connectionCB=NULL;
    exceptionCB=defaultExceptionCB;
    graphicalInfoCB=NULL;
    valueChangedCB=NULL;
    errHandler=errMsg;

    userData=NULL;
}

/***************************** taskInit *******************************/
int CCaObject2::taskInit(void)
{
  // CA Stuff
    int stat;

    stat=ca_task_initialize();
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::taskInit: "
	  "ca_task_initialize failed:\n%s\n",ca_message(stat));
	return 0;
    }
  
#ifdef todo
    stat=ca_add_fd_registration(::fdRegistrationICB,NULL);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::taskInit: "
	  "ca_add_fd_registration failed:\n%s\n",ca_message(stat));
	return 0;
    }
#endif

    stat=ca_add_exception_event(::exceptionICB,NULL);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::taskInit: "
	  "ca_add_exception_event failed:\n%s\n",ca_message(stat));
	return 0;
    }

    caStarted=1;
    return 1;
}

/***************************** taskExit *******************************/
int CCaObject2::taskExit(void)
{
  // CA Stuff
    int stat;

    stat=ca_task_exit();
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::taskExit: "
	  "ca_task_exit failed:\n%s\n",ca_message(stat));
	return 0;
    }
    caStarted=0;
  
    return 1;
}

// Always uses callback
/***************************** connect *******************************/
int CCaObject2::connect(const char *nameIn)
{
    int stat;
    int err=0;

  // Copy the name
    strncpy(name,nameIn,FULL_PVNAME_SZ);
    name[FULL_PVNAME_SZ-1]='\0';

  // Check if connected
    if(connected || Chid) disconnect();
	
  // Search
    stat=ca_search_and_connect(name,&Chid,::connectionICB,
      (void *)this);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::connect: "
	  "ca_search failed:\n%s\n",ca_message(stat)); err++;
	return 0;
    }

    ca_poll();

  // Success
  // (Chid has been allocated if ca_search_and_connect returns ECA_NORMAL)
    return 1;
}

/***************************** disconnect *****************************/
void CCaObject2::disconnect()
{
    int stat;

  // Stop monitoring (Remove value-changed callback)
    stopMonitor();

    if (Chid != NULL) {
#if 0
      // KE: This is not necessary according to Jeff
      // Remove the connection-changed callback
	stat=ca_change_connection_event(Chid,NULL);
	if(stat != ECA_NORMAL) {
	    ERRMSG("CCaObject2::disconnect: "
	      "ca_change_connection_event failed:\n%s\n",ca_message(stat));
	}
	
      // Replace the access rights callback
	stat=ca_replace_access_rights_event(Chid,NULL);
	if(stat != ECA_NORMAL) {
	    ERRMSG("CCaObject2::disconnect: "
	      "ca_replace_access_rights_event:\n%s\n",ca_message(stat));

	}
#endif

      // Clear the channel
	stat=ca_clear_channel(Chid);
	if(stat != ECA_NORMAL) {
	    ERRMSG("CCaObject2::disconnect: "
	      "ca_clear_channel failed:\n%s\n",ca_message(stat));

	}
    }
    Chid=NULL;
    connected=0;

  // Free the data
    if(data) delete [] data;
    data=NULL;

  // Process requests
    ca_poll();
}

// The startMonitor and stopMonitor routines essentialy start and stop
// the valueChanged callback and nothing else.  The graphicalInfo
// callback continues to be called on connection changes
/***************************** startMonitor ***************************/
int CCaObject2::startMonitor()
{
    int stat;

  // Check if connected
    if(!connected) {
	monitored=monitor_req;
	return 2;
    }

  // Check if monitored
    if(monitored == monitor_on) {
	stopMonitor();
    }

  // Add value-changed callback
#ifdef __USING_TIME_STAMP__
    stat=ca_add_array_event(
      dbf_type_to_DBR_TIME(ca_field_type(Chid)),
      ca_element_count(Chid),Chid,
      ::valueChangedICB,this,0.0,0.0,0.0,&Evid);
#else
    stat=ca_add_array_event(
      dbf_type_to_DBR_STS(ca_field_type(Chid)),
      ca_element_count(Chid),Chid,
      ::valueChangedICB,this,0.0,0.0,0.0,&Evid);
#endif
    if(stat != ECA_NORMAL) {
      // Set the pointer to NULL in case CA didn't.  We don't
      // want to use it or try to clear it later.
	Evid=NULL;
	ERRMSG("CCaObject2::startMonitor: "
	  "ca_add_array_event failed [%s]:\n %s\n",
	  ca_name(Chid)?ca_name(Chid):"Unknown",
	  ca_message(stat));
	return 0;
    }

  // Poll (Perhaps not necessary)
    ca_poll();

  // Success (assumed if using ca_poll)
    monitored=monitor_on;
    return 1;
}

/***************************** stopMonitor ****************************/
int CCaObject2::stopMonitor(void)
{
    int stat;

  // Check if monitored
    if(monitored != monitor_on || !Evid) {
	monitored=monitor_off;
	return 1;
    }

    stat=ca_clear_event(Evid);
    Evid=NULL;
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::stopMonitor: "
	  "ca_clear_event failed [%s]:\n %s\n",
	  ca_name(Chid)?ca_name(Chid):"Unknown",
	  ca_message(stat));
	return 0;
    }

  // Success (assumed)
    monitored=monitor_off;
    return 1;
}

// getCallback issues a ca_array_get_callback using the valueChanged
// callback.
/***************************** getCallback ****************************/
int CCaObject2::getCallback(void)
{
    int stat;

  // Check if we think it is connected
    if(!connected) {
	ERRMSG("CCaObject2::getCallback: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	ERRMSG("CCaObject2::getCallback: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the get with callback, pass this as the user data, use the
  // valueChanged callback
#ifdef __USING_TIME_STAMP__
    stat=ca_array_get_callback(
      dbf_type_to_DBR_TIME(ca_field_type(Chid)),
      ca_element_count(Chid),Chid,
      ::valueChangedICB,(void *)this);
#else
    stat=ca_array_get_callback(
      dbf_type_to_DBR_STS(ca_field_type(Chid)),
      ca_element_count(Chid),Chid,
      ::valueChangedICB,(void *)this);
#endif
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::getCallback: "
	  "ca_array_get_callback:\n%s\n",ca_message(stat));
	return 0;
    }
    ca_poll();

  // Success
    return 1;
}

// getGraphicalInfoCallback issues a ca_array_get_callback using the
// graphicalInfo callback.
/***************************** getGraphicalInfoCallback ***************/
int CCaObject2::getGraphicalInfoCallback(void)
{
    int stat;

  // Check if we think it is connected
    if(!connected) {
	ERRMSG("CCaObject2::getGraphicalInfoCallback: "
	  "Channel is not connected.\n");
	return 2;
    }

  // Check if CA thinks it is connected
    if(ca_state(Chid) != cs_conn) {
	ERRMSG("CCaObject2::getGraphicalInfoCallback: "
	  "Channel is not connected now.\n");
	return 3;
    }
    
  // Do the get with callback, pass this as the user data, use the
  // graphicalInfo callback
    stat=ca_array_get_callback(
      dbf_type_to_DBR_CTRL(ca_field_type(Chid)),
      ca_element_count(Chid),Chid,
      ::graphicalInfoICB,(void *)this);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::getGraphicalInfoCallback: "
	  "ca_array_get_callback:\n%s\n",ca_message(stat));
	return 0;
    }
    ca_poll();

  // Success
    return 1;
}

void CCaObject2::putDouble(double val)
{
    int stat;
    
    stat=ca_put(DBR_DOUBLE,Chid,&val);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::putDouble: ca_put failed: %s\n",
	  ca_message(stat));
    }
    ca_flush_io();
}

void CCaObject2::putLong(long lval)
{
    int stat;
    
    stat=ca_put(DBR_LONG,Chid,&lval);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::putLong: ca_put failed: %s\n",
	  ca_message(stat));
    }
    ca_flush_io();
}

void CCaObject2::putCharacterArray(char *chars, unsigned long size)
{
    int stat;
    
    stat=ca_array_put(DBR_CHAR,size,Chid,chars);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::putCharacterArray: ca_put failed: %s\n",
	  ca_message(stat));
    }
    ca_flush_io();
}

void CCaObject2::putString(char *string)
{
    int stat;
    
    stat=ca_put(DBR_STRING,Chid,string);
    if(stat != ECA_NORMAL) {
	ERRMSG("CCaObject2::putString: ca_put failed: %s\n",
	  ca_message(stat));
    }
    ca_flush_io();
}

int CCaObject2::getStringValue(char *string, int size, short prec,
  StringType type) const
{
    int isNumber=0;

    if(!string || size < 1) return 0;

  // If prec is < 0, use precision
    if(prec < 0) prec=precision;
  // Check values
    if(prec < 0) prec=0;
    if(prec > 17) prec=17;

    string[0]='\0';
    switch(dataType) {
    case DBF_STRING:
	if(array) {
	    strncpy(string,(char *)array,size);
	    string[size-1]='\0';
	}
	isNumber=FALSE;
	break;
    case DBF_ENUM:
	if(graphicalInfoAvailable && precision >= 0) {
	    int i=(int)value;
	    if(i >= 0 && i < (int)hopr+1){
		strncpy(string,stateStrings[i],size);
		string[size-1]='\0';
	    }
	    isNumber=FALSE;
	} else {
	    isNumber=TRUE;
	}
	break;
    case DBF_CHAR:
	if(type == STRING_STRING) {
	    if(array) {
		int len=MIN(elementCount+1,size);
		strncpy(string,(char *)array,len);
		string[len-1]='\0';
	    }
	    isNumber=FALSE;
	    break;
	}
    case DBF_INT:
    case DBF_LONG:
    case DBF_FLOAT:
    case DBF_DOUBLE:
	isNumber=TRUE;
	break;
    default:
	ERRMSG("CCaObject2::getStringValue: Unknown dataType [%d] for %s",
	  dataType,name);
	break;
    }
    
  // Do the value
    if(isNumber) {
	switch(type) {
	case STRING_STRING:
	    cvtDoubleToString(value,string,precision);
	    break;
	case DECIMAL_STRING:
	    cvtDoubleToString(value,string,precision);
#if 0
	  // KE: Don't do this, it can overflow the stack for large numbers
	  // Could be an exponential
	    if(strchr(string,'e')) {
		localCvtDoubleToString((double)value,string,precision);
	    }
#endif		    
	    break;
	case EXPONENTIAL_STRING:
#if 0
	    cvtDoubleToExpString(value,string,precision);
#endif
	  // KE: This is different from textEntryDRAW/valueToString
	    sprintf(string,"%.*e",precision,value);
	    break;
	case ENGR_NOTATION_STRING:
	    localCvtDoubleToExpNotationString(value,string,precision);
	    break;
	case COMPACT_STRING:
	    cvtDoubleToCompactString(value,string,precision);
	    break;
	case TRUNCATED_STRING:
	    cvtLongToString((long)value,string);
	    break;
	case HEXADECIMAL_STRING:
	    localCvtLongToHexString((long)value, string);
	    break;
	case OCTAL_STRING:
	    cvtLongToOctalString((long)value, string);
	    break;
	default:
	    ERRMSG("CCaObject2::getStringValue: Unknown StringType [%d] for %s",
	      type,name);
	    break;
	}
    }

    return 1;
}

// Returns 0.0 for a DBF_STRING and DBL_MAX for error
double CCaObject2::getArrayValueAsDouble(int i) const
{
    if(!array) return DBL_MAX;
    if(i < 0 || i >= elementCount) return DBL_MAX;

  // Branch depending on native type
    switch(this->dataType) {
    case DBF_STRING:
    {
	return 0.0;
    }
    case DBF_INT:
    {
	short *pShort=(short *)array;
	return (double)pShort[i];
    }
    case DBF_FLOAT:
    {
	float *pFloat=(float *)array;
	return (double)pFloat[i];
    }
    case DBF_ENUM:
    {
	unsigned short *pUShort=(unsigned short *)array;
	return (double)pUShort[i];
    }
    case DBF_CHAR:
    {
	unsigned char *pUChar=(unsigned char *)array;
	return (double)pUChar[i];
    }
    case DBF_LONG:
    {
	dbr_long_t *pLong=(dbr_long_t *)array;
	return (double)pLong[i];
    }
    case DBF_DOUBLE:
    {
	double *pDouble=(double *)array;
	return pDouble[i];
    }
    default:
	return DBL_MAX;
    }
}

// Uses static storage.  The user must copy this value if he wants to
// save it.
char *CCaObject2::getTimeStampString(void)
{
    static char timeStr[26];
#ifndef UNDER_CE
    time_t clock;
    
  // Convert the seconds part of the timestamp to UNIX time
    clock=this->time.secPastEpoch+631152000ul;
    strcpy(timeStr,ctime(&clock));
    timeStr[24]='\0';
#else
    strcpy(timeStr,"Timestamp not available");
#endif
    return timeStr;
}

void CCaObject2::accessRightsICB(struct access_rights_handler_args args)
{
    CCaObject2 *chan=(CCaObject2 *)ca_puser(args.chid);
    int readAccessChangedToRead=FALSE;

  // Check for valid values
    if(!chan || !chan->Chid) {
	ERRMSG("CCaObject2::accessRightsICB: Invalid channel");
	return;
    }

  // Determine if the rights are changing from no read to read
    if(!chan->readAccess && ca_read_access(chan->Chid)) {
	readAccessChangedToRead=TRUE;
    }
  // Change the access rights in the Record
    chan->readAccess=ca_read_access(chan->Chid);
    chan->writeAccess=ca_write_access(chan->Chid);

  // Do a get as in connectCB.  The get will cause the graphicalInfoCB
  // to be called.  This is necessary for some objects if the access
  // rights are changing from no read to read
    if(chan->pvCurConnected && chan->pvPrevConnected
      && readAccessChangedToRead) {
	int stat=ca_array_get_callback(
	  dbf_type_to_DBR_CTRL(ca_field_type(args.chid)),
	  1, args.chid, ::graphicalInfoICB, chan);
	if(stat != ECA_NORMAL) {
	    ERRMSG("CCaObject2::accessRightsICB: %s\n",
	      ca_message(stat));
	}
    }

  // Call the users access rights callback
    if(chan->accessRightsCB) (*chan->accessRightsCB)(chan);
}

// This is called when the connection changes and is instantiated via
// ca_search_and_connect
void CCaObject2::connectionICB(struct connection_handler_args args)
{
    CCaObject2 *chan=(CCaObject2 *)ca_puser(args.chid);
    int stat;

  // Check for valid values
    if(!chan || !chan->Chid) {
	ERRMSG("CCaObject2::connectionICB: Invalid channel");
	return;
    }

  // Do a get every time a channel is connected or reconnected and has
  // read access.  The get will cause the graphical info callback to
  // be called
    if(args.op == CA_OP_CONN_UP && ca_read_access(chan->Chid)) {
	chan->connected=1;
	stat=ca_array_get_callback(
	  dbf_type_to_DBR_CTRL(ca_field_type(args.chid)),
	  1, args.chid, ::graphicalInfoICB, chan);
	if(stat != ECA_NORMAL) {
	    ERRMSG("CCaObject2::connectionICB: "
	      "ca_array_get_callback [%s]:\n %s\n",
	      ca_name(chan->Chid)?ca_name(chan->Chid):"Unknown",
	      ca_message(stat));
	}
    }

  // Handle four cases: pvCurConnected or not and previously connected
  // or not
    if(args.op == CA_OP_CONN_UP) {
      // Connected
	chan->connected=1;
	if(chan->pvPrevConnected == FALSE) {
	  // Connected and not previously connected
	  // Set these first so they will be right for the valueChangedCB
	    chan->elementCount=ca_element_count(chan->Chid);
	    chan->dataType=ca_field_type(args.chid);
	    chan->pvCurConnected=TRUE;
	  // Add the access-rights-change callback and the
	  // significant-change (value, alarm or severity)
	  // callback.
	    stat=ca_replace_access_rights_event(
	      chan->Chid,::accessRightsICB);
	    if(stat != ECA_NORMAL) {
		ERRMSG("CCaObject2::connectionICB: "
		  "ca_replace_access_rights_event [%s]: %s\n",
		  ca_name(chan->Chid)?ca_name(chan->Chid):"Unknown",
		  ca_message(stat));
	    }
	  // Start the monitor if requested but not started
	    if(chan->monitored == monitor_req) chan->startMonitor();
	  // Set this one last so ca_replace_access_rights_event can
	  // use the old value
	    chan->pvPrevConnected=TRUE;
	} else {
	  // Connected and previously connected
	    chan->pvCurConnected=TRUE;
	}
    } else {
      // Not connected
	chan->connected=0;
	if(chan->pvPrevConnected == FALSE) {
	  // Not connected and not previously connected
	  // Probably doesn't happen -- if CA can't connect, this
	  //   routine never gets called
	    chan->pvCurConnected=FALSE;
	} else {
	  // Not connected but previously connected
	    chan->pvCurConnected=FALSE;
	}
    }

  // Call the user's connection callback
    if(chan->connectionCB) (chan->connectionCB)(chan);
}


// This is called when an exception occurs.  It is instantiated in
// taskInit.
void CCaObject2::exceptionICB(struct exception_handler_args args)
{
    if(exceptionCB) (*exceptionCB)(args);
}

#ifdef todo
// This is called when there is fd activity.  It is instantiated in
// taskInit.
void CCaObject2::fdRegistrationICB(void *dummy, int fd, int condition)
{
    UNREFERENCED(dummy);
    UNREFERENCED(fd);
    UNREFERENCED(condition);
}
#endif

// This is called as a result of ca_array_get_callback from
// connectionICB (when connection changes) and from
// getGraphicalInfoCallback (whenever called by the user).  Also from
// accessRightsICB.  It fills in hopr, lopr, precision, and value, not
// array.
void CCaObject2::graphicalInfoICB(struct event_handler_args args)
{
    CCaObject2 *chan=(CCaObject2 *)ca_puser(args.chid);
    int nBytes;
    int i;

  // Check for valid values
  // Same as for valueChangedICB
  // Don't need to check for read access, checking ECA_NORMAL is enough
    if(args.status != ECA_NORMAL) return;
    if(!args.dbr) {
	ERRMSG("CCaObject2::graphicalInfoICB: Invalid data [%s]\n",
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }
    if(!chan || !chan->Chid) {
	ERRMSG("CCaObject2::graphicalInfoICB: Invalid channel information [%s]\n",
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }
    if(chan->Chid != args.chid) {
	ERRMSG("CCaObject2::graphicalInfoICB: chid from args [%x] "
	  "does not match chid from CCaObject2 [%x]\n"
	  "  [%s]\n",
	  args.chid,
	  chan->Chid,
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }

  // Copy the information to the Channel struct
    nBytes=dbr_size_n(args.type, args.count);
    memcpy((void *)&(chan->info),args.dbr,nBytes);

  // Handle hopr, lopr, and precision
    switch(ca_field_type(args.chid)) {
    case DBF_STRING:
	chan->value=0.0;
	chan->hopr=0.0;
	chan->lopr=0.0;
	chan->precision=0;
	break;
    case DBF_ENUM:
	chan->value=(double) chan->info.e.value;
	chan->hopr=(double) chan->info.e.no_str - 1.0;
	chan->lopr=0.0;
	chan->precision=0;
	for (i=0; i < chan->info.e.no_str; i++) { 
	    chan->stateStrings[i]=chan->info.e.strs[i];
	}
	break;
    case DBF_CHAR:
	chan->value=(double) chan->info.c.value;
	chan->hopr=(double) chan->info.c.upper_disp_limit;
	chan->lopr=(double) chan->info.c.lower_disp_limit;
	chan->precision=0;
	break;
    case DBF_INT:
	chan->value=(double) chan->info.i.value;
	chan->hopr=(double) chan->info.i.upper_disp_limit;
	chan->lopr=(double) chan->info.i.lower_disp_limit;
	chan->precision=0;
	break;
    case DBF_LONG:
	chan->value=(double) chan->info.l.value;
	chan->hopr=(double) chan->info.l.upper_disp_limit;
	chan->lopr=(double) chan->info.l.lower_disp_limit;
	chan->precision=0;
	break;
    case DBF_FLOAT:
	chan->value=(double) chan->info.f.value;
	chan->hopr=(double) chan->info.f.upper_disp_limit;
	chan->lopr=(double) chan->info.f.lower_disp_limit;
	chan->precision=chan->info.f.precision;
	break;
    case DBF_DOUBLE:
	chan->value=(double) chan->info.d.value;
	chan->hopr=(double) chan->info.d.upper_disp_limit;
	chan->lopr=(double) chan->info.d.lower_disp_limit;
	chan->precision=chan->info.f.precision;
	break;
    default:
	ERRMSG("CCaObject2::graphicalInfoICB: Unknown data type");
	return;
    }

  // Adjust the precision
    if(chan->precision < 0) {
	ERRMSG("CCaObject2::graphicalInfoICB: "
	  "pv=\"%s\" precision=%d\n",
	  ca_name(chan->Chid), chan->precision);
	chan->precision=0;
    } else if(chan->precision > 17) {
	ERRMSG("CCaObject2::graphicalInfoICB: "
	  "pv=\"%s\" precision=%d\n",
	  ca_name(chan->Chid), chan->precision);
	chan->precision=17;
    }

  // Flag that the graphical info is available
    chan->graphicalInfoAvailable=1;

  // Call the user's graphicalInfo callback
    if(chan->graphicalInfoCB) (chan->graphicalInfoCB)(chan);
}

// This is called as a result of ca_add_array_event.  It gets called
// when the value changes when monitoring is on and from getCallback
// (whenever called by the user).  It fills in value and array, not
// hopr, lopr, or precision.
void CCaObject2::valueChangedICB(struct event_handler_args args)
{
    CCaObject2 *chan=(CCaObject2 *)ca_puser(args.chid);
    int statusChanged=FALSE;
    int severityChanged=FALSE;
    int zeroTransition=FALSE;
    double value;
    int nBytes;

  // Check for valid values
  // Same as for graphicalInfoCb
  // Don't need to check for read access, checking ECA_NORMAL is enough
    if(args.status != ECA_NORMAL) return;
    if(!args.dbr) {
	ERRMSG("CCaObject2::valueChangedICB: Invalid data [%]s\n",
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }
    if(!chan || !chan->Chid) {
	ERRMSG("CCaObject2::valueChangedICB: Invalid channel information [%s]\n",
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }
    if(chan->Chid != args.chid) {
	ERRMSG("CCaObject2::valueChangedICB: chid from args [%x] "
	  "does not match chid from channel [%x]\n"
	  "  [%s]\n",
	  args.chid,
	  chan->Chid,
	  ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	return;
    }
    
  // Allocate space for the data
    nBytes=dbr_size_n(args.type, args.count);
    if(!chan->data) {
      // No previous data, allocate
	chan->data=new dataBuf[nBytes];
	if(!chan->data) {
	    chan->dataSize=0;
	    ERRMSG("CCaObject2::valueChangedICB: Memory allocation error [%s]\n",
	      ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	    return;
	}
	chan->dataSize=nBytes;
    } else if(chan->dataSize < nBytes) {
      // Previous data exists but is too small.  Free it and allocate
      // it again
	delete [] chan->data;
	chan->data=new dataBuf[nBytes];
	if(chan->data == NULL) {
	    chan->dataSize=0;
	    ERRMSG("CCaObject2::valueChangedICB: Memory reallocation error"
	      " [%s]\n",
	      ca_name(args.chid)?ca_name(args.chid):"Name Unknown");
	    return;
	}
	chan->dataSize=nBytes;
    }

#ifdef __USING_TIME_STAMP__
  // Copy the returned time stamp to the Record
    chan->time=((dataBuf *)args.dbr)->s.stamp;
#endif

    
  // Set the array pointer in the Record to point to the value member
  // of the appropriate struct stored in the Channel.  (The returned
  // data is not stored there yet)
    switch(ca_field_type(args.chid)) {
    case DBF_STRING:
	chan->array=(void *)chan->data->s.value;
	break;
    case DBF_ENUM:
	chan->array=(void *)&(chan->data->e.value);
	break;
    case DBF_CHAR:
	chan->array=(void *)&(chan->data->c.value);
	break;
    case DBF_INT:
	chan->array=(void *)&(chan->data->i.value);
	break;
    case DBF_LONG:
	chan->array=(void *)&(chan->data->l.value);
	break;
    case DBF_FLOAT:
	chan->array=(void *)&(chan->data->f.value);
	break;
    case DBF_DOUBLE:
	chan->array=(void *)&(chan->data->d.value);
	break;
    default:
	break;
    }
    
  // For strings and arrays copy the returned data to the Channel (The
  // space is allocated but the data is not stored otherwise)
    if(ca_field_type(args.chid) == DBF_STRING ||
      ca_element_count(args.chid) > 1) {
	memcpy((void *)chan->data,args.dbr,nBytes);
    }

  // Copy the value from the returned data to the Record
    switch(ca_field_type(args.chid)) {
    case DBF_STRING:
	value=0.0;
	break;
    case DBF_ENUM:
	value=(double)((dataBuf *)(args.dbr))->e.value;
	break;
    case DBF_CHAR:
	value=(double)((dataBuf *)(args.dbr))->c.value;
	break;
    case DBF_INT:
	value=(double)((dataBuf *)(args.dbr))->i.value;
	break;
    case DBF_LONG:
	value=(double)((dataBuf *)(args.dbr))->l.value;
	break;
    case DBF_FLOAT:
	value=(double)((dataBuf *)(args.dbr))->f.value;
	break;
    case DBF_DOUBLE:
	value=((dataBuf *)(args.dbr))->d.value;
	break;
    default:
	value=0.0;
	break;
    }
    
  // Mark zero to nonzero transition
    if(((value == 0.0) && (chan->value != 0.0)) ||
      ((value != 0.0) && (chan->value == 0.0)))
      zeroTransition=TRUE;

  // Mark status changed
    if(chan->status != ((dataBuf *)(args.dbr))->d.status) {
	chan->status=((dataBuf *)(args.dbr))->d.status;
	statusChanged=TRUE;
    }

  // Mark severity changed
    if(chan->severity != ((dataBuf *)(args.dbr))->d.severity) {
	chan->severity=((dataBuf *)(args.dbr))->d.severity;
	severityChanged=TRUE;
    }

  // Set the new value into the record
    chan->value=value;
    
  // Call the valueChanged callback if there is a monitored change
    if(chan->valueChangedCB) {
	if(chan->monitorValueChanged) {
	    (*chan->valueChangedCB)(chan);
	} else if(chan->monitorStatusChanged && statusChanged) {
	    (*chan->valueChangedCB)(chan);
	} else if(chan->monitorSeverityChanged && severityChanged) {
	    (*chan->valueChangedCB)(chan);
	} else if(chan->monitorZeroTransition && zeroTransition) {
	    (*chan->valueChangedCB)(chan);
	}
    }
}

/***************************** errMsg *********************************/
int CCaObject2::errMsg(const char *fmt, ...)
{
    static char string[2048];
    va_list vargs;
    int nchars=0;

    va_start(vargs,fmt);
    nchars+=vsprintf(string,fmt,vargs);
    va_end(vargs);

#ifdef WIN32
  //    MessageBox(NULL,string,"CaObject Error",MB_OK|MB_ICONERROR);
#else
    fprintf(stderr,"%s\n",string);
#endif

    return nchars;
}

/***************************** defaultExceptionCB *********************/
void CCaObject2::defaultExceptionCB(struct exception_handler_args args)
{
#define MAX_EXCEPTIONS 25    
    static int nexceptions=0;
    static int ended=0;

    if(ended) return;

  // Get the users error handler from args.usr, which is this
    CCaObject2 *chan=(CCaObject2 *)args.usr;
    CCaObject2ErrorHandler userErrMsg=NULL;

    if(chan) userErrMsg=chan->getErrorHandler();

  // Avoid error storms
    if(nexceptions++ > MAX_EXCEPTIONS) {
	ended=1;
	if(userErrMsg) userErrMsg("CCaObject2::exceptionCB: Channel Access Exception:\n"
	  "Too many exceptions [%d]\n"
	  "No more will be handled\n"
	  "Please fix the problem and restart the program",
	  MAX_EXCEPTIONS);
	ca_add_exception_event(NULL,NULL);
	return;
    }

  // Get the time
    char timeStampStr[16];
    time_t now;
    struct tm *tblock;
    ::time(&now);
    tblock=localtime(&now);
    strftime(timeStampStr,20,"%b %d %H:%M:%S",tblock);

    if(userErrMsg) userErrMsg("CCaObject2::exceptionCB: Channel Access Exception:\n"
      "  Time: %s\n"
      "  Channel Name: %s\n"
      "  Native Type: %s\n"
      "  Native Count: %hu\n"
      "  Access: %s%s\n"
      "  IOC: %s\n"
      "  Message: %s\n"
      "  Context: %s\n"
      "  Requested Type: %s\n"
      "  Requested Count: %ld\n"
      "  Source File: %s\n"
      "  Line number: %u\n",
      timeStampStr,
      args.chid?ca_name(args.chid):"Unavailable",
      args.chid?dbf_type_to_text(ca_field_type(args.chid)):"Unavailable",
      args.chid?ca_element_count(args.chid):0,
      args.chid?(ca_read_access(args.chid)?"R":""):"Unavailable",
      args.chid?(ca_write_access(args.chid)?"W":""):"",
      args.chid?ca_host_name(args.chid):"Unavailable",
      ca_message(args.stat)?ca_message(args.stat):"Unavailable",
      args.ctx?args.ctx:"Unavailable",
      dbf_type_is_valid(args.type)?dbf_type_to_text(args.type):"Invalid",
      args.count,
      args.pFile?args.pFile:"Unavailable",
      args.pFile?args.lineNo:0);
}

// extern "C" wrappers
extern "C" {

/***************************** accessRightsICB ************************/
static void accessRightsICB(struct access_rights_handler_args args)
{
    CCaObject2::accessRightsICB(args);
}

/***************************** connectionICB **************************/
static void connectionICB(struct connection_handler_args args)
{
    CCaObject2::connectionICB(args);
}

/***************************** exceptionICB ***************************/
static void exceptionICB(struct exception_handler_args args)
{
    CCaObject2::exceptionICB(args);
}

/***************************** graphicalInfoICB ***********************/
static void graphicalInfoICB(struct event_handler_args args)
{
    CCaObject2::graphicalInfoICB(args);
}

#ifdef todo
/***************************** fdRegistrationICB **********************/
static void fdRegistrationICB(void *dummy, int fd, int condition)
{
    CCaObject2::fdRegistrationICB(dummy,fd,condition);
}
#endif

/***************************** valueChangedICB ************************/
static void valueChangedICB(struct event_handler_args args)
{
    CCaObject2::valueChangedICB(args);
}

}  // End extern "C"

// Utilities

/***************************** localCvtDoubleToExpNotationString ******/
static void localCvtDoubleToExpNotationString(double value, char *textField,
  unsigned short precision)
{
    double absVal, newVal;
    int minus;
    int exp, k, l;
    char TF[MAX_CONVERT_SIZE];
    
  /* The PREC field is a short, which can be negative, which will be a
   *   large number when converted to unsigned short.  This is handled
   *   in localCvtDoubleToString. */

    absVal=fabs(value);
    minus=(value < 0.0 ? TRUE : FALSE);
    newVal=absVal;
    
    if (absVal < 1.) {
      /* absVal < 1. */
	exp=0;
	if (absVal != 0.) {     /* really ought to test against some epsilon */
	    do {
		newVal *= 1000.0;
		exp += 3;
	    } while (newVal < 1.);
	}
	localCvtDoubleToString(newVal,TF,precision);
	k=0; l=0;
	if (minus) textField[k++]='-';
	while (TF[l] != '\0') textField[k++]=TF[l++];
	textField[k++]='e';
	if (exp == 0) {
	    textField[k++]='+';     /* want e+00 for consistency with norms */
	} else {
	    textField[k++]='-';
	}
	textField[k++]='0' + exp/10;
	textField[k++]='0' + exp%10;
	textField[k++]='\0';
	
    } else {
	
      /* absVal >= 1. */
	exp=0;
	while (newVal >= 1000.) {
	    newVal *= 0.001; /* since multiplying is usually faster than dividing */
	    exp += 3;
	}
	localCvtDoubleToString(newVal,TF,precision);
	k=0; l=0;
	if (minus) textField[k++]='-';
	while (TF[l] != '\0') textField[k++]=TF[l++];
	textField[k++]='e';
	textField[k++]='+';
	textField[k++]='0' + exp/10;
	textField[k++]='0' + exp%10;
	textField[k++]='\0';
    }
}

/***************************** localCvtDoubleToHexString **************/
static int localCvtLongToHexString(long source, char  *pdest)
{
    static char hex_digit_to_ascii[16]={'0','1','2','3','4','5','6','7',
					'8','9','a','b','c','d','e','f'};
    long  val,temp;
    char  digit[10];
    int   i,j;
    char  *startAddr=pdest;
 
    *pdest++='0';
    *pdest++='x';
    if(source==0) {
	*pdest++='0';
    } else {
	val=source;
	temp=0xf & val;
	val=val >> 4;
	digit[0]=hex_digit_to_ascii[temp];
	if(val < 0) val=val & 0xfffffff;
	for(i=1; val!=0; i++) {
	    temp=0xf & val;
	    val=val >> 4;
	    digit[i]=hex_digit_to_ascii[temp];
	}
	for(j=i-1; j>=0; j--) {
	    *pdest++=digit[j];
	}

    }
    *pdest=0;
    return((int)(pdest-startAddr));
}

static void localCvtDoubleToString(double flt_value, char  *pstr_value,
  unsigned short precision)
{
  /* The PREC field is a short, which can be negative, which will be a
   *   large number when converted to unsigned short.  Handle this by
   *   not letting precision be larger than 17 (as for libCom:
   *   cvtDoubleToString()).  Letting it be zero here would be OK, but
   *   not in general, since precision is the number of significant
   *   digits for g and e formats. */
    if(precision > 17) precision=17;
    sprintf(pstr_value,"%.*f",(int)precision,flt_value);
}
